#include "lua.h"
#include "lauxlib.h"

#include <stdlib.h>
#include <readline/readline.h>

#define LCOMPLETERLIBNAME "rlcompleter"

static int initialized = 0;

static lua_State *storedL;

static char *
generator(const char *text, int state)
{
  int call_result;
  int height = lua_gettop(storedL);
  char *s;

  rl_filename_completion_desired = 0;
  rl_filename_quoting_desired = 0;

  lua_pushlightuserdata(storedL, (void *)generator);
  lua_gettable(storedL, LUA_REGISTRYINDEX);
  if (!lua_isfunction(storedL, -1)) {
    lua_settop(storedL, height);
    return NULL;
  }
  lua_pushstring(storedL, text);
  lua_pushnumber(storedL, state);
  call_result = lua_pcall(storedL, 2, 1, 0);
  if (call_result) {
    lua_settop(storedL, height);
    return NULL;
  } 
    
  if (lua_isnil(storedL, -1)) {
    lua_settop(storedL, height);
    return NULL;
  }
  if (!lua_isstring(storedL, -1)) {
    lua_settop(storedL, height);
    return NULL;
  }

  s = strdup(lua_tostring(storedL, -1));
  lua_settop(storedL, height);
  return s;
}

char **
mycompleter(const char *text, int start, int end)
{
  rl_attempted_completion_over = 1;
  return rl_completion_matches(text, generator);
}

#define BREAKCHARS " \t\n\"\\`~!@#$%^&*()-+=[]{}|;:,<>/?"

static void
initialize(void) 
{
  /* rl_initialize(); */
  rl_attempted_completion_function = mycompleter;
  /* someday maybe lua can set these, hence the strdup */
  rl_basic_word_break_characters = strdup(BREAKCHARS);
  rl_basic_word_break_characters = strdup(BREAKCHARS);
  rl_completion_append_character = '\0';

  initialized = 1;
}

static int setcompleter(lua_State *L) {
  if (!initialized) {
    initialize();
  }
  if (!lua_isfunction(L, 1)) {
    luaL_typerror(L, 1, "function");
  }
  lua_pushlightuserdata(L, (void *)generator);
  lua_pushvalue(L, 1);
  lua_settable(L, LUA_REGISTRYINDEX);

  return 0;
}


static int redisplay(lua_State *L) {
  if (!initialized) { initialize(); }

  rl_forced_update_display();
  return 0;
}

static const luaL_reg completerlib[] = {
  {"_set", setcompleter},
  {"redisplay", redisplay},
  {NULL, NULL}
};

int luaopen_rlcompleter(lua_State *L) {
  storedL = L;
  luaL_openlib(L, LCOMPLETERLIBNAME, completerlib, 0);
  return 0;
}
